import shlex
import urllib.parse
import uuid
from os import path
from uuid import UUID

import yaml
from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.db import models
from django.db.models import Q
from django.utils.functional import cached_property
from enumfields import Enum, EnumField
from rest_framework.exceptions import ValidationError

from arkindex.dataimport.providers import get_provider, git_providers
from arkindex.documents.models import ClassificationState, Element
from arkindex.project.aws import S3FileMixin, S3FileStatus
from arkindex.project.fields import ArrayField
from arkindex.project.models import IndexableModel
from ponos.models import Artifact, State, Workflow


class DataImportMode(Enum):
    Images = 'images'
    PDF = 'pdf'
    Repository = 'repository'
    Elements = 'elements'
    IIIF = 'iiif'
    Workers = 'workers'
    Transkribus = 'transkribus'


class DataImport(IndexableModel):
    """
    A single import workflow
    """

    name = models.CharField(null=True, blank=True, max_length=100)
    creator = models.ForeignKey('users.User', on_delete=models.CASCADE, related_name='imports')
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='imports', null=True)
    mode = EnumField(DataImportMode, max_length=30)
    files = models.ManyToManyField('dataimport.DataFile', related_name='imports')
    revision = models.ForeignKey(
        'dataimport.Revision', related_name='dataimports', on_delete=models.CASCADE, blank=True, null=True)
    workflow = models.OneToOneField('ponos.Workflow', on_delete=models.SET_NULL, null=True, blank=True)
    versions = models.ManyToManyField('dataimport.WorkerVersion', through='dataimport.WorkerRun', related_name='imports')

    # Used to define the root element on which the workflow will run
    element = models.ForeignKey('documents.Element', on_delete=models.SET_NULL, null=True, blank=True)

    # Used to define the element type during folder elements creation for Files/IIIF imports
    folder_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='folder_imports',
    )

    # Used to define the element type during non-folder elements creation for Files/IIIF imports,
    # or to filter elements by type with Elements imports
    element_type = models.ForeignKey(
        'documents.ElementType',
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='element_imports',
    )

    # Used to filter out elements with a name that doesn't contain the substring, only on Elements imports
    name_contains = models.CharField(null=True, blank=True, max_length=250)

    # Used to filter elements by best class, either having or not having a best class (boolean)
    # or having a specific best class (UUID), only on Elements imports
    best_class = models.CharField(null=True, blank=True, max_length=36)

    # Used to save a user's selection for Element workflows
    elements = models.ManyToManyField(
        'documents.Element',
        through='dataimport.DataImportElement',
        related_name='imports',
    )

    # Used to include sub-elements of the actual query
    load_children = models.BooleanField(default=False)

    # Used to define the collection ID and entities import for Transkribus import
    collection_id = models.PositiveIntegerField(null=True, blank=True)
    build_entities = models.BooleanField(default=False)

    class Meta:
        ordering = ['corpus', '-created']

    @property
    def state(self):
        if not self.workflow:
            return State.Unscheduled
        # This allows annotating a DataImport queryset with "last_run" and preventing duplicate SQL queries
        if hasattr(self, 'last_run'):
            # last_run may be None when there is a workflow without any tasks
            if self.last_run is None:
                return State.Unscheduled
            return self.workflow.get_state(self.last_run)
        else:
            return self.workflow.state

    def _get_filters(self):
        filters = {
            "corpus_id": self.corpus_id,
        }
        if self.name_contains:
            filters['name__contains'] = self.name_contains
        if self.element_type:
            filters['type_id'] = self.element_type_id
        else:
            # Limit the scope of types available to merge
            # This prevent memory from exploding when no type is selected
            filters['type__corpus_id'] = self.corpus_id

        return filters

    def _get_classifications_filters(self):
        if self.best_class is None:
            return

        # Generic ORM query to find best classes:
        # - elements with a validated classification
        # - OR where high confidence is True
        best_classifications = Q(classifications__state=ClassificationState.Validated) \
            | Q(classifications__high_confidence=True)

        # List elements without any best classes, by inverting the query above
        if self.best_class in ('false', '0'):
            return ~best_classifications

        try:
            # Filter on a specific class
            class_filter = UUID(self.best_class)
            return best_classifications & Q(classifications__ml_class_id=class_filter)
        except (TypeError, ValueError):
            # By default, use all best classifications
            return best_classifications

    def list_elements(self):
        """
        Return a queryset of elements involved in this process
        """
        if self.mode not in (DataImportMode.Elements, DataImportMode.Workers):
            return Element.objects.none()

        elements = None

        if self.element:
            # Assert the element has the same corpus as the process
            assert self.element.corpus_id == self.corpus_id, \
                f'Element {self.element.id} is not part of corpus {self.corpus_id}'
            # Handle a single element
            elements = Element.objects.filter(id=self.element.id)

        elif self.elements.exists():
            # Handle a selection of elements
            elements = self.elements.all()
            # Check all elements are in the same corpus as the process
            corpus = list(elements.values_list('corpus_id', flat=True).distinct())
            assert len(corpus) == 1 and corpus[0] == self.corpus_id, \
                f'Some elements on this process are not part of corpus {self.corpus_id}'

        if elements is not None and self.load_children:
            # Load all the children elements whose path contains the pre-selected elements
            # Those children are appended to the pre-selection
            elements |= Element.objects.filter(
                paths__path__overlap=map(str, elements.values_list('id', flat=True)),
            )

        # Load the full corpus, only when elements has not been populated before
        if elements is None:
            # Handle all elements of the process corpus
            elements = Element.objects.filter(corpus=self.corpus_id)

        # Filter elements depending on process properties
        elements = elements.filter(**self._get_filters())
        class_filters = self._get_classifications_filters()
        if class_filters is not None:
            # Distinct is required because multiple classes may match the filter
            elements = elements.filter(class_filters).distinct()

        return elements

    def build_workflow(self, chunks=None, thumbnails=False, corpus_id=None):
        '''
        Create a ponos workflow with a recipe according to configuration
        '''
        ml_workflow_chunks = 1
        import_task_name = 'import'
        if self.mode == DataImportMode.Elements:
            import_task_name = 'initialisation'

        assert self.workflow is None, 'A workflow is already setup'
        if self.mode == DataImportMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')

        if self.mode == DataImportMode.Repository:
            assert self.revision is not None, \
                'A revision is required to create an import workflow from GitLab repository'
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': (
                        f"python -m arkindex_tasks.import_git {self.revision.id}"
                        + (f' --corpus-id {corpus_id}' if corpus_id else '')
                    )
                },
            }

        elif self.mode in (DataImportMode.Elements, DataImportMode.Workers):
            if chunks is not None:
                assert isinstance(chunks, int) and chunks > 0, 'Chunks count should be an positive integer'
                assert chunks < 11, 'Import distribution is limited to 10 chunks'
                ml_workflow_chunks = chunks

            args = [
                'python', '-m', 'arkindex_tasks.init_elements', str(self.id),
                '--chunks-number', str(ml_workflow_chunks),
            ]

            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': ' '.join(args),
                }
            }

        elif self.mode == DataImportMode.IIIF:
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python -m arkindex_tasks.import_iiif.dataimport {}'.format(self.id),
                },
            }

        elif self.mode == DataImportMode.Transkribus:
            args = [
                'python', '-m', 'arkindex_tasks.import_transkribus',
                str(self.collection_id),
                '--corpus', str(self.corpus.id),
            ]
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': ' '.join(args),
                }
            }

        else:
            tasks = {
                import_task_name: {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python -m arkindex_tasks.import_files {}'.format(self.id),
                },
            }

        # Import entities directy after import step
        if self.mode == DataImportMode.Transkribus and self.build_entities:
            transcriptions_path = shlex.quote(path.join('/data', import_task_name, str(self.collection_id), 'transcriptions.json'))
            tasks['build_entities'] = {
                'image': settings.ARKINDEX_TASKS_IMAGE,
                'command': 'python -m arkindex_tasks.build_entities {}'.format(transcriptions_path),
                'parents': [import_task_name],
            }

        elts_chunk_files = ['elements.json']
        if ml_workflow_chunks > 1:
            elts_chunk_files = ['elements_chunk_{}.json'.format(n) for n in range(1, ml_workflow_chunks + 1)]

        for index, elts_chunk in enumerate(elts_chunk_files, start=1):
            # Add a name suffix if task is distributed
            task_suffix = ''
            if len(elts_chunk_files) > 1:
                task_suffix = '_{}'.format(index)

            elements_path = shlex.quote(path.join('/data', import_task_name, elts_chunk))

            # Generate thumbnails directly after import step
            if thumbnails:
                tasks['thumbnails{}'.format(task_suffix)] = {
                    'image': settings.ARKINDEX_TASKS_IMAGE,
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails {}'.format(elements_path),
                    'parents': [import_task_name],
                }

            # Generate a task for each WorkerRun on the DataImport
            for worker_run in self.worker_runs.using('default').all():
                task_name = f'{worker_run.version.slug}{task_suffix}'
                tasks[task_name] = worker_run.build_task_recipe(import_task_name, elements_path, suffix=task_suffix)

        # Build the workflow in db
        recipe = settings.PONOS_RECIPE.copy()
        recipe['tasks'] = tasks
        if 'env' not in recipe:
            recipe['env'] = {}
        recipe['env'].update({
            'ARKINDEX_PROCESS_ID': str(self.id),
        })
        if self.mode == DataImportMode.Transkribus:
            recipe['env']['TRANSKRIBUS_EMAIL'] = settings.TRANSKRIBUS_EMAIL
            recipe['env']['TRANSKRIBUS_PASSWORD'] = settings.TRANSKRIBUS_PASSWORD
            if settings.TRANSKRIBUS_WORKER_VERSION:
                recipe['env']['TRANSKRIBUS_WORKER_VERSION'] = str(settings.TRANSKRIBUS_WORKER_VERSION)
        elif settings.IMPORTS_WORKER_VERSION and self.mode in (DataImportMode.Images, DataImportMode.PDF):
            recipe['env']['WORKER_VERSION_ID'] = str(settings.IMPORTS_WORKER_VERSION)
        return Workflow.objects.create(recipe=yaml.dump(recipe))

    def start(self, chunks=None, thumbnails=False, corpus_id=None):
        self.workflow = self.build_workflow(chunks, thumbnails, corpus_id)
        self.save()
        self.workflow.start()

    def retry(self):
        if self.mode == DataImportMode.Repository and self.revision is not None and not self.revision.repo.enabled:
            raise ValidationError('Git repository does not have any valid credentials')
        if self.workflow:
            self.workflow.retry()
        else:
            self.start()


class DataImportElement(models.Model):
    """
    Link between DataImports and Elements to allow workflows from selections.
    May be removed later once the ElementPath algorithms are optimized,
    as a selection could then be turned into a set.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    dataimport = models.ForeignKey(DataImport, on_delete=models.CASCADE, related_name='dataimport_elements')
    element = models.ForeignKey('documents.Element', on_delete=models.CASCADE, related_name='dataimport_elements')

    class Meta:
        unique_together = (
            ('dataimport', 'element'),
        )


class DataFile(S3FileMixin, models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    size = models.PositiveIntegerField(help_text='file size in bytes')
    content_type = models.CharField(max_length=120)
    corpus = models.ForeignKey('documents.Corpus', on_delete=models.CASCADE, related_name='files')
    status = EnumField(S3FileStatus, default=S3FileStatus.Unchecked, max_length=50)

    s3_bucket = settings.AWS_STAGING_BUCKET

    @cached_property
    def s3_key(self):
        return str(self.id)

    class Meta:
        ordering = ['corpus', 'name']

    def perform_check(self, save=True, raise_exc=False):
        """
        Check Datafile's existence, content type and size from S3
        """
        try:
            assert self.exists(), 'File has not been uploaded'
            assert self.s3_object.content_length == self.size, \
                f'Uploaded file size is {self.s3_object.content_length} bytes, expected {self.size} bytes'
            assert self.s3_object.content_type == self.content_type, \
                f'Uploaded file content type is {self.s3_object.content_type}, expected {self.content_type}'
            self.status = S3FileStatus.Checked
        except AssertionError:
            self.status = S3FileStatus.Error
            if raise_exc:
                raise
        finally:
            if save:
                self.save()


class RepositoryType(Enum):
    IIIF = 'iiif'
    Worker = 'worker'


class Repository(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    url = models.URLField(unique=True)
    type = EnumField(RepositoryType)
    hook_token = models.CharField(max_length=250, unique=True)
    credentials = models.ForeignKey(
        'users.OAuthCredentials', on_delete=models.SET_NULL, related_name='repos', blank=True, null=True)
    provider_name = models.CharField(
        max_length=50,
        choices=[(p.__name__, p.display_name) for p in git_providers],
    )
    git_ref_revisions = models.ManyToManyField('dataimport.Revision', through='dataimport.GitRef')
    memberships = GenericRelation('users.Right', 'content_id')

    class Meta:
        verbose_name_plural = 'repositories'

    @property
    def provider_class(self):
        if not self.provider_name:
            raise ValueError("Empty provider_name")
        return get_provider(self.provider_name)

    @property
    def provider(self):
        if self.provider_class is None:
            raise NotImplementedError(f"Missing provider {self.provider_name}")
        return self.provider_class(credentials=self.credentials)

    @property
    def enabled(self):
        return self.credentials is not None and self.provider_class is not None

    def __str__(self):
        return f'{self.url}'


class GitRefType(Enum):
    Branch = 'branch'
    Tag = 'tag'


class GitRef(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    type = EnumField(GitRefType)
    name = models.CharField(max_length=250)
    repository = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='refs')
    revision = models.ForeignKey('dataimport.Revision', on_delete=models.CASCADE, related_name='refs')

    class Meta:
        unique_together = (('name', 'repository'), )


class Revision(IndexableModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    repo = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='revisions')
    hash = models.CharField(max_length=50)
    message = models.TextField()
    author = models.CharField(max_length=50)

    class Meta:
        unique_together = (('repo', 'hash'), )

    @property
    def commit_url(self):
        return '{}/commit/{}'.format(self.repo.url.rstrip('/'), self.hash)

    def __str__(self):
        message = ' "{}"'.format(self.message.splitlines()[0]) if self.message else ''
        return '{}{} by {}'.format(self.hash[:8], message, self.author)


class Worker(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=100)
    slug = models.CharField(max_length=100)
    type = models.CharField(max_length=50)
    repository = models.ForeignKey('dataimport.Repository', on_delete=models.CASCADE, related_name='workers')
    memberships = GenericRelation('users.Right', 'content_id')

    # Allow any authentified user to see and execute this worker
    public = models.BooleanField(default=False)

    class Meta:
        unique_together = (('slug', 'repository'),)

    def __str__(self):
        return f'{self.name}'


class WorkerVersionState(Enum):
    Created = 'created'
    Processing = 'processing'
    Available = 'available'
    Error = 'error'


class WorkerVersion(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    worker = models.ForeignKey('dataimport.Worker', on_delete=models.CASCADE, related_name='versions')
    revision = models.ForeignKey('dataimport.Revision', on_delete=models.CASCADE, related_name='versions')
    configuration = models.JSONField()
    state = EnumField(WorkerVersionState, default=WorkerVersionState.Created)

    # Foreign key towards the ponos Artifact holding the image content
    docker_image = models.ForeignKey(Artifact, on_delete=models.SET_NULL, null=True, blank=True)

    # The Docker internal image id (sha256:xxx) that can be shared across multiple images
    docker_image_iid = models.CharField(null=True, blank=True, max_length=80)

    class Meta:
        unique_together = (('worker', 'revision'),)
        constraints = [
            models.CheckConstraint(
                check=~models.Q(state=WorkerVersionState.Available, docker_image_id=None),
                name='workerversion_available_requires_docker_image',
            )
        ]

    def __str__(self):
        return f'{self.worker} for revision {self.revision}'

    @property
    def docker_image_name(self) -> str:
        parsed_url = urllib.parse.urlparse(self.worker.repository.url)
        return f'{parsed_url.netloc}{parsed_url.path}/{self.worker.slug}:{self.id}'.lower()

    @property
    def slug(self):
        truncated_hash = str(self.id)[0:6]
        return f'{self.worker.slug}_{truncated_hash}'

    @property
    def docker_command(self):
        if 'docker' not in self.configuration:
            return
        return self.configuration['docker'].get('command')


class WorkerRun(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    dataimport = models.ForeignKey('dataimport.DataImport', on_delete=models.CASCADE, related_name='worker_runs')
    version = models.ForeignKey('dataimport.WorkerVersion', on_delete=models.CASCADE, related_name='worker_runs')
    parents = ArrayField(models.UUIDField())

    class Meta:
        unique_together = (('version', 'dataimport'),)

    def build_task_recipe(self, import_task_name, elements_path, suffix=''):
        '''
        Build recipe dict for the Task that will represent this WorkerRun in ponos using :
        - the docker image name given by the WorkerVersion
        - the command given by the WorkerVersion
        - the artifact given by the WorkerVersion that will help us download the distant Docker image (S3)
        - a list of parent tasks slugs, using the base import task as a default
        '''
        if self.parents:
            parents = [
                f'{worker_run.version.slug}{suffix}'
                for worker_run in WorkerRun.objects.using('default').filter(id__in=self.parents)
            ]
        else:
            parents = [import_task_name]

        return {
            'image': self.version.docker_image_iid or self.version.docker_image_name,
            'command': self.version.docker_command,
            'artifact': str(self.version.docker_image.id),
            'parents': parents,
            'env': {"TASK_ELEMENTS": elements_path, "WORKER_VERSION_ID": str(self.version.id)}
        }


class WorkerActivityState(Enum):
    Queued = 'queued'
    Started = 'started'
    Processed = 'processed'
    Error = 'error'


class WorkerActivity(IndexableModel):
    """
    Many-to-many relationship between Element and WorkerVersion
    Used to track the activity of a worker version among mutliple elements
    """
    # Using an UUID helps to execute SQL raw INSERT
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    element = models.ForeignKey(
        'documents.Element',
        on_delete=models.CASCADE,
        related_name='activities',
    )
    worker_version = models.ForeignKey(
        WorkerVersion,
        on_delete=models.CASCADE,
        related_name='activities',
    )
    state = EnumField(
        WorkerActivityState,
        default=WorkerActivityState.Queued
    )

    class Meta:
        unique_together = (
            ('worker_version', 'element'),
        )
