import base64
import logging
import urllib.parse
import uuid
from abc import ABC, abstractmethod

import yaml
from django.conf import settings
from django.urls import reverse
from gitlab import Gitlab, GitlabCreateError, GitlabGetError
from rest_framework.exceptions import APIException, AuthenticationFailed, NotAuthenticated, ValidationError
from yaml.parser import ParserError

logger = logging.getLogger(__name__)

CONF_PATH = settings.WORKERS_CONF_PATH
CONF_VERSION = settings.WORKERS_CONF_VERSION


class GitProvider(ABC):

    display_name = None

    def __init__(self, credentials=None, url=None):
        if credentials:
            from arkindex.users.models import OAuthCredentials
            assert isinstance(credentials, OAuthCredentials)
            self.url = credentials.provider_url
        if url:
            self.url = url
        self.credentials = credentials

    @abstractmethod
    def list_repos(self, query=None):
        """
        List all repositories or filter with a search query.
        """

    @abstractmethod
    def get_latest_commit_sha(self, repo):
        """
        Retrieve latest commit sha from the default branch on a given repository.
        """

    @abstractmethod
    def create_repo(self, **kwargs):
        """
        Create a Repository instance from an external repository
        """

    @abstractmethod
    def create_hook(self, repository, project_id=None, request=None, base_url=None):
        """Create a webhook to receive events from the provide"""

    @abstractmethod
    def get_repository_type(self, project):
        """ Retrieve a project repository type """

    def update_or_create_ref(self, repo, revision, name, type):
        """
        Update or create a GitRef on a given repository:
        - If a GitRef with given name already exists on the given repository, we only update the
        linked revision with the given one
        - Else a GitRef is created on the given repository thanks to given name, type and revision
        """
        from arkindex.dataimport.models import GitRef
        try:
            ref = repo.refs.get(name=name)
            ref.revision = revision
            ref.save()
        except GitRef.DoesNotExist:
            repo.refs.create(name=name, type=type, revision=revision)

    def get_or_create_revision(self, repo, sha, save=True):
        from arkindex.dataimport.models import Revision
        try:
            return self.get_revision(repo, sha), False
        except Revision.DoesNotExist:
            return self.create_revision(repo, sha, save=save), True

    def get_revision(self, repo, sha):
        return repo.revisions.get(hash=sha)

    @abstractmethod
    def create_revision(self, repo, sha, save=True):
        """
        Create a Revision instance for a given commit hash of a given repository.
        """

    @abstractmethod
    def handle_webhook(self, repo, request):
        """
        Handle a webhook event on a given repository.
        """

    @abstractmethod
    def to_dict(self, project):
        """
        Returns a dict specifying id, name and url from a repository
        """

    def start_imports(self, rev):
        """
        Create and start processes from a repository revision
        For a worker repository, start a single process to build the new workers
        For other repositories, start a process for each corpus depending on the repository
        """
        from arkindex.dataimport.models import DataImport, RepositoryType, DataImportMode
        mode = DataImportMode.Repository
        user = rev.repo.credentials.user
        refs = ", ".join(rev.refs.values_list('name', flat=True))
        if not refs:
            refs = rev.hash
        name = f"Import {refs} from {rev.repo.url}"

        if rev.repo.type == RepositoryType.Worker:
            # Build revision workers in a new process
            dataimport = rev.dataimports.create(creator=user, mode=mode, name=name)
            dataimport.start(thumbnails=False)
            return

        # Otherwise start an import for each corpora linked to this repository
        data_imports = DataImport.objects.bulk_create([
            DataImport(
                revision=rev,
                creator=user,
                mode=mode,
                corpus=corpus,
                name=name,
            ) for corpus in rev.repo.corpora.all()
        ])
        for di in data_imports:
            di.start(thumbnails=True, corpus_id=di.corpus.id)


class GitLabProvider(GitProvider):

    display_name = "GitLab"
    url = 'https://gitlab.com'

    def _try_get_project(self, gl, id):
        try:
            return gl.projects.get(id)
        except GitlabGetError as e:
            raise APIException("Error while fetching GitLab project: {}".format(str(e)))

    def _get_project_from_repo(self, repo):
        assert repo.credentials, "Missing Gitlab credentials"

        gl = Gitlab(self.url, oauth_token=repo.credentials.token)
        return self._try_get_project(gl, urllib.parse.urlsplit(repo.url).path.strip('/'))

    def list_repos(self, query=None):
        if not self.credentials:
            raise NotAuthenticated()
        gl = Gitlab(self.url, oauth_token=self.credentials.token)
        # Creating a webhook on a repo requires Maintainer (40) or Owner (50) access levels
        # See https://docs.gitlab.com/ce/api/members.html#valid-access-levels
        return gl.projects.list(min_access_level=40, search=query)

    def get_latest_commit_sha(self, repo):
        project = self._get_project_from_repo(repo)
        # Since ref_name isn't specified, we will use the repository default branch
        # See : https://docs.gitlab.com/ee/api/commits.html#list-repository-commits
        return project.commits.list()[0].id

    def get_repository_type(self, project):
        """
        Retrieve arkindex repository type from the Gitlab project repository configuration
        """
        from arkindex.dataimport.models import RepositoryType
        conf_file = next((file for file in project.repository_tree() if file['name'] == CONF_PATH), None)
        if not conf_file:
            raise ValidationError(f'{project.name} project is missing a {CONF_PATH} configuration file')

        file_data = project.repository_raw_blob(conf_file['id'])
        try:
            conf = yaml.safe_load(file_data)
            assert isinstance(conf, dict)
        except (ParserError, AssertionError):
            raise ValidationError(f'{project.name} project has an invalid {CONF_PATH} configuration file')

        errors = {}
        if str(conf.get('version')) != CONF_VERSION:
            errors['version'] = f'version is not defined or is different from the latest version ({CONF_VERSION})'
        try:
            conf_type = RepositoryType(conf.get('type'))
        except ValueError:
            errors['type'] = f'type is not defined or does not match handled repository types ({list(RepositoryType)})'
        if errors:
            raise ValidationError(f'{project.name} project {CONF_PATH} configuration file could not be parsed', errors)
        return conf_type

    def to_dict(self, project):
        return {
            'id': project.id,
            'name': project.name_with_namespace,
            'url': project.web_url
        }

    def create_repo(self, id=None, corpus=None, request=None, **kwargs):
        if not self.credentials and request:
            raise NotAuthenticated()
        gl = Gitlab(self.url, oauth_token=self.credentials.token)
        project = self._try_get_project(gl, int(id))

        from arkindex.dataimport.models import Repository
        if Repository.objects.filter(url=project.web_url).exists():
            raise ValidationError("A repository with this URL already exists")

        # Determine the user's access level on this project
        # When it is inherited from the group and not overriden in the project, project_access is not defined.
        # When the project isn't in a group, or the user is added to a specific project in a group,
        # group_access is not defined. project_access overrides group_access.
        access_level = 0
        if project.permissions.get('group_access'):
            access_level = project.permissions['group_access']['access_level']
        if project.permissions.get('project_access'):
            access_level = project.permissions['project_access']['access_level']

        # Maintainer level (40) is required
        if access_level < 40:
            raise ValidationError("Maintainer or Owner access is required to add a GitLab repository")

        repo_type = self.get_repository_type(project)

        repo = self.credentials.repos.create(
            type=repo_type,
            url=project.web_url,
            hook_token=base64.b64encode(uuid.uuid4().bytes).decode('utf-8'),
            provider_name=self.__class__.__name__,
        )

        # Create a webhook using informations from the HTTP request
        self.create_hook(repo, project_id=int(id), request=request)

        return repo

    def create_hook(self, repository, project_id=None, request=None, base_url=None):
        """Configure the Gitlab hook to get events for that project"""

        # Load project using a project if or its repo path
        gitlab = Gitlab(self.url, oauth_token=repository.credentials.token)
        if project_id:
            project = self._try_get_project(gitlab, project_id)
        else:
            path = urllib.parse.urlparse(repository.url).path[1:]
            project = self._try_get_project(gitlab, path)

        try:
            url = reverse('api:import-hook', kwargs={'pk': repository.id})
            if settings.BACKEND_PUBLIC_URL_OAUTH:
                url = urllib.parse.urljoin(settings.BACKEND_PUBLIC_URL_OAUTH, url)
            elif request is not None:
                url = request.build_absolute_uri(url)
            elif base_url is not None:
                url = urllib.parse.urljoin(base_url, url)
            else:
                raise Exception("You need to specify an HTTP request or a base url")
            logger.info(f"Webhook will be created as {url}")

            # Delete already configured hooks to be able to update
            for hook in project.hooks.list(all=True):
                if hook.url == url:
                    hook.delete()
                    logger.info(f"Deleted existing hook {hook.id}")

            # Create a new hook
            hook = project.hooks.create({
                'url': url,
                'push_events': True,
                "tag_push_events": True,
                'token': repository.hook_token,
            })
            logger.info(f"Created new hook {hook.id}")
        except GitlabCreateError as e:
            raise APIException("Error while creating GitLab hook: {}".format(str(e)))

    def create_revision(self, repo, sha, save=True):
        from arkindex.dataimport.models import Revision, GitRefType

        project = self._get_project_from_repo(repo)
        commit = project.commits.get(sha)

        rev = Revision(
            repo=repo,
            hash=sha,
            message=commit.message,
            author=commit.author_name,
        )
        if save:
            rev.save()
            for ref in commit.refs():
                try:
                    ref_type = GitRefType(ref['type'])
                    self.update_or_create_ref(repo, rev, ref['name'], ref_type)
                except ValueError:
                    logger.warning(f'Git reference with type {ref["type"]} was ignored during revision creation')
                    continue

        return rev

    def update_revision_references(self, repo, sha):
        """
        Update all references for a specific revision
        This is needed when a tag is pushed after its commit
        is already ingested
        """
        from arkindex.dataimport.models import GitRefType

        project = self._get_project_from_repo(repo)
        commit = project.commits.get(sha)

        rev = repo.revisions.get(hash=sha)
        for ref in commit.refs():
            try:
                ref_type = GitRefType(ref['type'])
                self.update_or_create_ref(repo, rev, ref['name'], ref_type)
            except ValueError:
                logger.warning(f'Git reference with type {ref["type"]} was ignored during revision creation')
                continue

    def update_repository_references(self, repo):
        """
        List all available references (branches and tags) on a remote repository
        and create new references or update existing ones
        New import process are started for new references
        """
        from arkindex.dataimport.models import GitRefType
        project = self._get_project_from_repo(repo)

        def _update_reference(ref, ref_type):
            assert isinstance(ref_type, GitRefType)
            try:
                rev, created = repo.revisions.get_or_create(
                    hash=ref.commit['id'],
                    defaults={
                        'message': ref.commit['message'],
                        'author': ref.commit['author_email'],
                    }
                )
                self.update_or_create_ref(repo, rev, ref.name, ref_type)
                if created:
                    logger.info(f"Starting import for {ref_type.value} {ref.name}")
                    self.start_imports(rev)
            except ValueError:
                logger.warning(f'{ref_type.value} {ref.name} was ignored during revision creation')

        for branch in project.branches.list():
            _update_reference(branch, GitRefType.Branch)

        for tag in project.tags.list():
            _update_reference(tag, GitRefType.Tag)

    def handle_webhook(self, repo, request):
        from arkindex.dataimport.models import GitRefType

        if 'HTTP_X_GITLAB_EVENT' not in request.META:
            raise ValidationError("Missing GitLab event type")

        if request.META['HTTP_X_GITLAB_EVENT'] not in ('Push Hook', 'Tag Push Hook'):
            raise ValidationError("Unsupported GitLab event type")

        if 'HTTP_X_GITLAB_TOKEN' not in request.META:
            raise NotAuthenticated("Missing GitLab secret token")
        if request.META['HTTP_X_GITLAB_TOKEN'] != repo.hook_token:
            raise AuthenticationFailed("Invalid GitLab secret token")

        if not isinstance(request.data, dict) or 'object_kind' not in request.data:
            raise ValidationError('Bad payload format')

        kind = request.data.get('object_kind')
        if kind not in ('push', 'tag_push'):
            raise ValidationError("Unsupported GitLab event type")

        sha = request.data.get('checkout_sha')
        if kind == 'tag_push':
            if sha:
                # When a commit SHA is present, we need to add a reference on an existing commit
                self.update_revision_references(repo, sha)

            else:
                # If there isn't any SHA it means that a tag was deleted
                ref = request.data.get('ref')
                if not ref:
                    raise ValidationError('Missing tag reference')

                # Delete existing tag
                tag_name = ref[10:] if ref.startswith('refs/tags/') else ref
                repo.refs.filter(name=tag_name, type=GitRefType.Tag).delete()

            return

        if not sha:
            raise ValidationError('Missing checkout SHA')

        # Already took care of this event
        if repo.revisions.filter(hash=sha).exists():
            return

        rev = self.create_revision(repo, sha)

        self.start_imports(rev)


git_providers = [
    GitLabProvider,
]
oauth_to_git = {
    "gitlab": GitLabProvider,
}


def get_provider(name):
    return next(filter(lambda p: p.__name__ == name, git_providers), None)


def from_oauth(name):
    return oauth_to_git.get(name)
