from unittest.mock import MagicMock, patch

from django.urls import reverse
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.serializers import DateTimeField

from arkindex.dataimport.models import DataImport, DataImportMode, Repository, RepositoryType
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import Role, User
from ponos.models import State, Workflow

RECIPE = '''
tasks:
  first:
    image: alpine
'''


class TestRepositories(FixtureTestCase):

    def setUp(self):
        super().setUp()
        self.creds = self.user.credentials.get()
        self.iiif_repo = self.creds.repos.get(type=RepositoryType.IIIF)
        self.worker_repo = self.creds.repos.get(type=RepositoryType.Worker)
        self.rev = self.iiif_repo.revisions.get()
        self.serialized_revision = {
            'id': str(self.rev.id),
            'repo_id': str(self.iiif_repo.id),
            'author': self.rev.author,
            'commit_url': self.rev.commit_url,
            'date': DateTimeField().to_representation(value=self.rev.created),
            'hash': self.rev.hash,
            'message': self.rev.message
        }

    def build_repository_response(self):
        return [
            {
                'id': str(self.iiif_repo.id),
                'type': RepositoryType.IIIF.value,
                'corpora': [str(c.id) for c in self.iiif_repo.corpora.all()],
                'enabled': True,
                'git_clone_url': None,
                'url': self.iiif_repo.url,
                'workers': [],
                'authorized_users': self.iiif_repo.memberships.count(),
            }, {
                'id': str(self.worker_repo.id),
                'type': RepositoryType.Worker.value,
                'corpora': [str(c.id) for c in self.worker_repo.corpora.all()],
                'enabled': True,
                'git_clone_url': None,
                'url': self.worker_repo.url,
                'workers': [
                    {
                        'id': str(w.id),
                        'name': w.name,
                        'type': w.type,
                        'slug': w.slug,
                        'repository_id': str(self.worker_repo.id),
                    } for w in self.worker_repo.workers.all()
                ],
                'authorized_users': self.worker_repo.memberships.count(),
            }
        ]

    def test_delete_credentials_null(self):
        """
        Check deleting OAuthCredentials do not delete repositories and revisions
        """
        self.creds.delete()
        self.assertTrue(Repository.objects.filter(url='http://gitlab/repo').exists())
        self.iiif_repo.refresh_from_db()
        self.assertTrue(self.iiif_repo.revisions.exists())

    def test_no_credentials_no_import(self):
        """
        Check Repository imports do not start without credentials
        """
        self.iiif_repo.credentials = None
        self.iiif_repo.save()
        workflows_count = Workflow.objects.count()

        di = DataImport.objects.create(
            mode=DataImportMode.Repository,
            revision=self.rev,
            creator=self.superuser,
            corpus=self.corpus,
        )

        with self.assertRaises(ValidationError):
            di.start()

        with self.assertRaises(ValidationError):
            di.retry()

        self.assertEqual(Workflow.objects.count(), workflows_count)

    def test_list_repository_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:repository-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_repository_external_user(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:repository-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertCountEqual(data['results'], [])

    def test_list_repository_user_readable(self):
        self.iiif_repo.memberships.create(user=self.user, level=Role.Admin.value)
        self.worker_repo.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:repository-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 2)
        self.assertCountEqual(data['results'], self.build_repository_response())

    def test_list_repository_internal_user(self):
        """
        An internal user may access the full list of repositories.
        Multiple repository serialization should not include the git_clone_url field.
        """
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:repository-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 2)
        self.assertCountEqual(data['results'], self.build_repository_response())

    def test_list_repository_multiple_refs(self):
        """
        Listed repositories should not be duplicated
        """
        self.iiif_repo.corpora.create()
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:repository-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 2)
        self.assertCountEqual(data['results'], self.build_repository_response())

    def test_repository_retrieve_requires_login(self):
        response = self.client.get(
            reverse('api:repository-retrieve', kwargs={'pk': str(self.iiif_repo.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_repository_retrieve_wrong_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:repository-retrieve', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_repository_retrieve_no_right(self):
        self.client.force_login(self.user)
        self.assertFalse(self.user.is_internal)
        self.assertFalse(self.user.is_admin)
        response = self.client.get(
            reverse('api:repository-retrieve', kwargs={'pk': str(self.iiif_repo.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_repository_retrieve(self):
        """
        An user with an access right on a repository is able to retrieve it without its clone URL
        """
        self.iiif_repo.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:repository-retrieve', kwargs={'pk': str(self.iiif_repo.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            'id': str(self.iiif_repo.id),
            'type': RepositoryType.IIIF.value,
            'corpora': [str(c.id) for c in self.iiif_repo.corpora.all()],
            'enabled': True,
            'git_clone_url': None,
            'url': self.iiif_repo.url,
            'workers': [],
            'authorized_users': self.iiif_repo.memberships.count(),
        })

    def test_repository_retrieve_internal(self):
        """
        An internal user is able to retrieve any repository with its clone URL
        """
        self.client.force_login(self.internal_user)
        response = self.client.get(
            reverse('api:repository-retrieve', kwargs={'pk': str(self.iiif_repo.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            'id': str(self.iiif_repo.id),
            'type': RepositoryType.IIIF.value,
            'corpora': [str(c.id) for c in self.iiif_repo.corpora.all()],
            'enabled': True,
            'git_clone_url': 'https://oauth2:oauth-token@gitlab/repo',
            'url': self.iiif_repo.url,
            'workers': [],
            'authorized_users': self.iiif_repo.memberships.count(),
        })

    def test_repository_retrieve_disabled_repo(self):
        self.iiif_repo.credentials = None
        self.iiif_repo.save()
        self.iiif_repo.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:repository-retrieve', kwargs={'pk': str(self.iiif_repo.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, {
            'id': str(self.iiif_repo.id),
            'type': RepositoryType.IIIF.value,
            'corpora': [str(c.id) for c in self.iiif_repo.corpora.all()],
            'enabled': False,
            'git_clone_url': None,
            'url': self.iiif_repo.url,
            'workers': [],
            'authorized_users': self.iiif_repo.memberships.count(),
        })

    def test_repository_delete_not_admin(self):
        """
        A user cannot delete a repository if he has no admin access, even if the repository is based on its credentials
        """
        creator = User.objects.create()
        repo = Repository.objects.create(type=RepositoryType.Worker, credentials=creator.credentials.create())
        repo.memberships.create(user=creator, level=Role.Contributor.value)
        self.client.force_login(creator)
        response = self.client.delete(reverse('api:repository-retrieve', kwargs={'pk': str(repo.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_repository_delete_creator_write_access(self):
        """
        An admin member of a repository can delete it
        """
        self.iiif_repo.memberships.create(user=self.user, level=Role.Admin.value)
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:repository-retrieve', kwargs={'pk': str(self.iiif_repo.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Repository.DoesNotExist):
            self.iiif_repo.refresh_from_db()

    def test_revision_retrieve_requires_login(self):
        response = self.client.get(
            reverse('api:revision-retrieve', kwargs={'pk': str(self.rev.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_revision_retrieve_wrong_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:revision-retrieve', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_revision_retrieve_read_right(self):
        """
        A user with no right or a read right is not allowed to retrieve a revision
        """
        self.iiif_repo.memberships.create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:revision-retrieve', kwargs={'pk': str(self.rev.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_revision_retrieve_contributor_right(self):
        """
        A user with a contributor access to the revision repo is allowed to retrieve information
        """
        self.iiif_repo.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:revision-retrieve', kwargs={'pk': str(self.rev.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertDictEqual(data, self.serialized_revision)

    @patch('arkindex.users.models.OAuthCredentials.git_provider_class')
    def test_list_repositories(self, git_provider_mock):
        repos = [MagicMock(id=i) for i in range(1, 4)]
        git_provider_mock().list_repos.return_value = repos
        git_provider_mock().to_dict = lambda repo: {
            'id': repo.id,
            'name': f'Repository {repo.id}',
            'url': f'http://repo_{repo.id}'
        }

        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:available-repositories', kwargs={'pk': self.creds.id})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), [
            {'id': 1, 'name': 'Repository 1', 'url': 'http://repo_1'},
            {'id': 2, 'name': 'Repository 2', 'url': 'http://repo_2'},
            {'id': 3, 'name': 'Repository 3', 'url': 'http://repo_3'}
        ])

    def test_list_repositories_wrong_creds(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:available-repositories', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_repositories_create_id_required(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:available-repositories', kwargs={'pk': self.creds.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'id': ['This field is required.']})

    @patch('arkindex.users.models.OAuthCredentials.git_provider_class')
    @patch('arkindex.dataimport.models.DataImport.start')
    def test_list_repositories_create_workers_repo(self, dataimport_start_mock, git_provider_mock):
        """
        Importing a workers repository creates a workers process
        """
        self.client.force_login(self.user)
        git_provider_mock().get_or_create_revision.return_value = self.rev, None
        repo = Repository.objects.create(type=RepositoryType.Worker)
        git_provider_mock().create_repo.return_value = repo
        with self.assertNumQueries(10):
            response = self.client.post(reverse('api:available-repositories', kwargs={'pk': self.creds.id}), {'id': 1111})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data_import = DataImport.objects.get(revision=self.rev)
        self.assertEqual(response.json(), {
            'name': None,
            'id': str(data_import.id),
            'corpus': None,
            'element': None,
            'element_type': None,
            'load_children': False,
            'name_contains': None,
            'files': [],
            'folder_type': None,
            'mode': DataImportMode.Repository.value,
            'revision': self.serialized_revision,
            'state': State.Unscheduled.value,
            'workflow': None
        })

        # Thumbnails are not generated for Workers imports
        self.assertEqual(dataimport_start_mock.call_count, 1)
        _, kwargs = dataimport_start_mock.call_args
        self.assertDictEqual(kwargs, {'thumbnails': False})
        self.assertIsNone(data_import.corpus)

        # User is granted an admin access to the repository
        repo_right = repo.memberships.get(user=self.user)
        self.assertTrue(repo_right.level >= Role.Admin.value)

    @patch('arkindex.users.models.OAuthCredentials.git_provider_class')
    @patch('arkindex.dataimport.models.DataImport.start')
    def test_list_repositories_create_iiif_repo(self, dataimport_start_mock, git_provider_mock):
        """
        Importing a IIIF repository result in the creation of a new corpus for the data import
        """
        self.client.force_login(self.user)
        git_provider_mock().get_or_create_revision.return_value = self.rev, None
        git_provider_mock().create_repo.return_value = self.iiif_repo
        with self.assertNumQueries(14):
            response = self.client.post(reverse('api:available-repositories', kwargs={'pk': self.creds.id}), {'id': 1111})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data_import = DataImport.objects.get(revision=self.rev)
        import_corpus = data_import.corpus
        self.assertEqual(response.json(), {
            'name': None,
            'id': str(data_import.id),
            'corpus': str(import_corpus.id),
            'element': None,
            'element_type': None,
            'load_children': False,
            'name_contains': None,
            'files': [],
            'folder_type': None,
            'mode': DataImportMode.Repository.value,
            'revision': self.serialized_revision,
            'state': State.Unscheduled.value,
            'workflow': None
        })

        # Thumbnails are generated for IIIF imports
        self.assertEqual(dataimport_start_mock.call_count, 1)
        _, kwargs = dataimport_start_mock.call_args
        self.assertDictEqual(kwargs, {'thumbnails': True, 'corpus_id': import_corpus.id})

        self.assertFalse(import_corpus.public)
        self.assertEqual(import_corpus.name, 'IIIF import from http://gitlab/repo')
        self.assertEqual(import_corpus.types.count(), 11)

        # User is granted an admin access to both the repository and the corpus
        corpus_right = import_corpus.memberships.get(user=self.user)
        self.assertTrue(corpus_right.level >= Role.Admin.value)

        repo_right = import_corpus.repository.memberships.get(user=self.user)
        self.assertTrue(repo_right.level >= Role.Admin.value)
