import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import Repository, RepositoryType, Revision, Worker, WorkerVersion, WorkerVersionState
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Right, Role, User
from ponos.models import Workflow

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''


class TestWorkersWorkerVersions(FixtureAPITestCase):
    """
    Test workers and worker versions endpoints
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(type=RepositoryType.Worker)
        # User is the creator of the repository
        cls.repo.memberships.create(user=cls.user, level=Role.Admin.value)
        cls.rev = cls.repo.revisions.get()

        cls.worker_1 = Worker.objects.get(slug='reco')
        cls.worker_2 = Worker.objects.get(slug='dla')
        cls.rev2 = Revision.objects.create(
            hash='1234',
            message='commit message',
            author='Bob',
            repo=cls.repo,
        )

    def setUp(self):
        self.version_1 = WorkerVersion.objects.get(worker=self.worker_1)

    # Tests on get_query_set for WorkerList
    def test_workers_list_requires_login(self):
        response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_list(self):
        """
        User is able to list workers on the repository as he has an admin access
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 2,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(self.worker_2.id),
                    'repository_id': str(self.repo.id),
                    'name': 'Document layout analyser',
                    'slug': 'dla',
                    'type': 'dla',
                }, {
                    'id': str(self.worker_1.id),
                    'repository_id': str(self.repo.id),
                    'name': 'Recognizer',
                    'slug': 'reco',
                    'type': 'recognizer',
                }
            ]
        })

    def test_workers_list_requires_contributor(self):
        """
        User is not able to list workers with a guest access
        """
        self.repo.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 0,
            'next': None,
            'number': 1,
            'previous': None,
            'results': []
        })

    def test_workers_list_public_attribute(self):
        """
        Any authenticated user is able to list a worker with the public attribute
        """
        user = User.objects.create_user('user42@test.fr', 'abcd')
        self.worker_1.public = True
        self.worker_1.save()
        self.client.force_login(user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [{
                'id': str(self.worker_1.id),
                'repository_id': str(self.repo.id),
                'name': 'Recognizer',
                'slug': 'reco',
                'type': 'recognizer',
            }]
        })

    def test_workers_list_name_filter(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:workers-list'), {'name': 'layout'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(self.worker_2.id),
                    'repository_id': str(self.worker_2.repository.id),
                    'name': 'Document layout analyser',
                    'slug': 'dla',
                    'type': 'dla',
                }
            ]
        })

    def test_workers_list_filter_repository(self):
        repo2 = Repository.objects.create(
            url='http://gitlab/repo2',
            type=RepositoryType.Worker,
            hook_token='hook-token2',
            credentials=self.creds,
            provider_name='GitLabProvider'
        )
        worker_2 = Worker.objects.create(
            repository=repo2,
            name='Worker 2',
            slug='worker_2',
            type='classifier'
        )
        worker_2.memberships.create(user=self.user, level=Role.Contributor.value)

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:workers-list'), {'repository_id': str(repo2.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [
                {
                    'id': str(worker_2.id),
                    'repository_id': str(repo2.id),
                    'name': 'Worker 2',
                    'slug': 'worker_2',
                    'type': 'classifier'
                }
            ]
        })

    def test_workers_retrieve_required_login(self):
        response = self.client.get(reverse('api:worker-retrieve', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workers_retrieve(self):
        """
        A user is able to retrieve a worker if he has a guest access on it or its repository
        """
        repo2 = Repository.objects.create(
            url='http://gitlab/repo2',
            type=RepositoryType.Worker,
            hook_token='hook-token2',
            credentials=self.creds,
            provider_name='GitLabProvider'
        )
        worker_2 = repo2.workers.create(name='Worker 2', slug='worker_2', type='classifier')
        repo2.memberships.create(user=self.user, level=Role.Guest.value)

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:worker-retrieve', kwargs={'pk': str(worker_2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(worker_2.id),
            'repository_id': str(repo2.id),
            'name': worker_2.name,
            'slug': worker_2.slug,
            'type': worker_2.type,
        })

    def test_worker_create_requires_login(self):
        response = self.client.post(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_workers_retrieve_no_rights(self):
        new_user = User.objects.create(email='new@test.fr', verified_email=True)
        self.client.force_login(new_user)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:worker-retrieve', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'You do not have a guest access to this worker.'})

    def test_worker_create_repository_not_uid(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:workers-list'),
            data={
                'repository_id': 'repo',
                'name': 'Worker post',
                'slug': 'worker_post',
                'type': 'classifier'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'repository_id': ['Must be a valid UUID.']})

    def test_workers_create_unexisting_repository(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:workers-list'),
            data={
                'repository_id': uuid.uuid4(),
                'name': 'Worker post',
                'slug': 'worker_post',
                'type': 'classifier'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_workers_create_required_fields(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'repository_id': ['This field is required.'],
            'name': ['This field is required.'],
            'slug': ['This field is required.'],
            'type': ['This field is required.']
        })

    def test_workers_create_worker(self):
        """
        User with an admin right on a repository is able to create a worker
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:workers-list'),
            data={
                'repository_id': str(self.repo.id),
                'name': 'Worker post',
                'slug': 'worker_post',
                'type': 'classifier'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertNotEqual(data['id'], str(self.worker_1.id))
        self.assertEqual(data['name'], 'Worker post')
        self.assertEqual(data['slug'], 'worker_post')
        self.assertEqual(data['type'], 'classifier')

    def test_workers_create_internal_user(self):
        """
        Internal users (e.g. workers) are allowed to create a worker even if they have no defined right
        """
        self.assertEqual(
            Right.objects.filter(
                content_id__in=(self.repo.id, *self.repo.workers.values_list('id', flat=True)),
                user=self.internal_user
            ).count(),
            0
        )
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:workers-list'),
            data={
                'repository_id': str(self.repo.id),
                'name': 'Worker post',
                'slug': 'worker_post',
                'type': 'classifier'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_workers_post_return_existing_worker(self):
        self.client.force_login(self.user)
        # A worker named Worker 1 already exists with the same slug on this repo
        response = self.client.post(
            reverse('api:workers-list'),
            data={
                'repository_id': str(self.repo.id),
                'name': 'Worker Test',
                'slug': 'reco',
                'type': 'classifier'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.worker_1.id))
        self.assertEqual(data['name'], 'Recognizer')
        self.assertEqual(data['slug'], 'reco')
        self.assertEqual(data['type'], 'recognizer')

    # Tests on get_query_set for WorkerVersionList
    def test_versions_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_versions_list(self):
        self.client.force_login(self.user)
        last_revision = self.repo.revisions.create(hash='42', message='A second version', author='Michel')
        last_version = self.worker_1.versions.create(
            revision=last_revision,
            configuration={'a': 'bc'},
            state=WorkerVersionState.Error,
            docker_image=self.version_1.docker_image
        )
        with self.assertNumQueries(14):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 2)
        # The revision created in this test has a recent timestamp so it is displayed first
        first_version = data['results'][0]
        revision = first_version.pop('revision')
        self.assertEqual(revision['id'], str(last_revision.id))
        self.assertEqual(revision['message'], 'A second version')
        self.assertDictEqual(first_version, {
            'id': str(last_version.id),
            'configuration': {'a': 'bc'},
            'docker_image': str(self.version_1.docker_image.id),
            'docker_image_iid': None,
            'docker_image_name': f'my_repo.fake/workers/worker/reco:{last_version.id}',
            'state': 'error',
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
            }
        })

    def test_workers_versions_list_requires_contributor(self):
        """
        User is not able to list worker versions with a guest access
        """
        self.repo.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:workers-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 0,
            'next': None,
            'number': 1,
            'previous': None,
            'results': []
        })

    def test_versions_list_public_worker(self):
        """
        Any authenticated user is able to list versions for public workers
        """
        user = User.objects.create_user('user42@test.fr', 'abcd')
        self.worker_1.public = True
        self.worker_1.save()
        self.client.force_login(user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(self.worker_1.versions.first().id))

    def test_versions_list_filter_worker(self):
        self.client.force_login(self.user)
        worker_2 = Worker.objects.create(
            repository=self.repo,
            name='Worker 2',
            slug='worker_2',
            type='classifier'
        )
        version_2 = WorkerVersion.objects.create(
            worker=worker_2,
            revision=self.rev,
            configuration={"test": "test2"}
        )

        with self.assertNumQueries(13):
            response = self.client.get(reverse('api:worker-versions', kwargs={'pk': str(worker_2.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        version = data['results'][0]
        self.assertEqual(version['id'], str(version_2.id))

    def test_versions_post_requires_internal(self):
        """
        Only internal user are able to create new versions
        """
        access = [
            (None, status.HTTP_403_FORBIDDEN),
            (self.user, status.HTTP_403_FORBIDDEN),
            (self.superuser, status.HTTP_403_FORBIDDEN),
            (self.internal_user, status.HTTP_201_CREATED)
        ]
        for user, status_code in access:
            if user:
                self.client.force_login(user)
            response = self.client.post(
                reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
                data={'revision': str(self.rev2.id), 'configuration': {"test": "test2"}}, format='json'
            )
            self.assertEqual(response.status_code, status_code)

    def test_versions_post_non_existing_worker(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
            data={'revision': str(self.rev2.id), 'configuration': {"test": "test2"}}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_versions_post_available_requires_docker_image(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
            data={
                'revision': str(self.rev2.id),
                'configuration': {"test": "test2"},
                'state': 'available',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_versions_post_return_existing_worker_version(self):
        self.client.force_login(self.internal_user)
        # A worker version already exists for this worker and this revision
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
            data={'revision': str(self.rev.id), 'configuration': {"test": "test1"}}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": 42})
        self.assertEqual(data['revision']['id'], str(self.rev.id))
        self.assertEqual(data['state'], 'available')

    def test_versions_post_empty(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'revision': ['This field is required.'],
            'configuration': ['This field is required.']
        })

    def test_versions_post_unrelated_revision(self):
        """
        It is not possible to create a version between a worker and a revision of two different repositories
        """
        rev = Revision.objects.create(
            hash='1337',
            message='The IIIF import I am admin on',
            author='Yann',
            repo=self.creds.repos.get(type=RepositoryType.IIIF)
        )
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
            data={'revision': str(rev.id), 'configuration': {"test": "test2"}}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'revision': ['The revision must be part of the same repository as the worker.']
        })

    def test_versions_post_create_worker_version(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:worker-versions', kwargs={'pk': str(self.worker_1.id)}),
            data={'revision': str(self.rev2.id), 'configuration': {"test": "test2"}}, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertNotEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['revision']['id'], str(self.rev2.id))
        self.assertEqual(data['state'], 'created')

    def test_retrieve_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.get(
            reverse('api:version-retrieve', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_version_no_login(self):
        """
        Worker versions may be retrieved with no authentication in order
        to see the version on public ressources
        """
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        revision = data.pop('revision')
        self.assertEqual(revision['id'], str(self.rev.id))
        self.assertDictEqual(data, {
            'id': str(self.version_1.id),
            'configuration': {"test": 42},
            'docker_image': str(self.version_1.docker_image.id),
            'docker_image_iid': None,
            'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
            'state': 'available',
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
            }
        })

    def test_retrieve_version(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        revision = data.pop('revision')
        self.assertEqual(revision['id'], str(self.rev.id))
        self.assertDictEqual(data, {
            'id': str(self.version_1.id),
            'configuration': {"test": 42},
            'docker_image': str(self.version_1.docker_image.id),
            'docker_image_iid': None,
            'docker_image_name': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
            'state': 'available',
            'worker': {
                'id': str(self.worker_1.id),
                'name': self.worker_1.name,
                'type': self.worker_1.type,
            }
        })

    def test_update_version_requires_login(self):
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': '12341234-1234-1234-1234-123412341234',
                'state': 'error'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_update_version_invalid_id(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': '12341234-1234-1234-1234-123412341234',
                'state': 'error'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_worker_version_non_internal(self):
        """
        Only internal users are allowed to update a worker version
        """
        docker_image = self.version_1.docker_image
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': str(docker_image.id),
                'state': 'error'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'Only internal users are allowed to update a worker version.'})

    def test_update_worker_version(self):
        """
        Update worker version artifact, configuration and state
        """
        docker_image = self.version_1.docker_image
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': str(docker_image.id),
                'state': 'error'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.version_1.refresh_from_db()
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['docker_image'], str(docker_image.id))
        self.assertIsNone(data['docker_image_iid'])
        self.assertEqual(data['state'], 'error')

    def test_update_version_valid(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        workflow = Workflow.objects.create(recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        artifact = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )

        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={
                'configuration': {"test": "test2"},
                'docker_image': str(artifact.id),
                'docker_image_iid': 'sha256:deadbeef',
                'state': 'available'
            }, format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.version_1.id))
        self.assertEqual(data['configuration'], {"test": "test2"})
        self.assertEqual(data['docker_image'], str(artifact.id))
        self.assertEqual(data['docker_image_iid'], 'sha256:deadbeef')
        self.assertEqual(data['state'], 'available')

    def test_update_version_available_requires_docker_image(self):
        self.version_1.state = WorkerVersionState.Created
        self.version_1.docker_image = None
        self.version_1.save()

        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'state': 'available'},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_update_version_available_docker_image_not_null(self):
        workflow = Workflow.objects.create(recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        self.version_1.docker_image = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )
        self.version_1.state = WorkerVersionState.Available
        self.version_1.save()

        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'docker_image': None},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_update_version_available_docker_image_iid_not_null(self):
        workflow = Workflow.objects.create(recipe=RECIPE)
        workflow.start()
        task = workflow.tasks.get(slug='first')
        self.version_1.docker_image = task.artifacts.create(
            path='path/to/file.json',
            size=100,
            content_type='application/json',
        )
        self.version_1.docker_image_iid = 'something:latest'
        self.version_1.state = WorkerVersionState.Available
        self.version_1.save()

        self.client.force_login(self.internal_user)
        response = self.client.patch(
            reverse('api:version-retrieve', kwargs={'pk': str(self.version_1.id)}),
            data={'docker_image_iid': None},
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': [
                'A WorkerVersion must have a Docker image artifact and a Docker image ID set when it is available.'
            ]
        })

    def test_property_docker_image_name(self):
        self.assertEqual(self.version_1.docker_image_name, f'my_repo.fake/workers/worker/reco:{self.version_1.id}')

    def test_property_docker_image_name_explicit(self):
        version_2 = WorkerVersion.objects.create(
            id="12341234-1234-1234-1234-123412341234",
            worker=self.worker_1,
            revision=self.rev2,
            configuration={"test": "test1"}
        )
        self.assertEqual(
            version_2.docker_image_name,
            'my_repo.fake/workers/worker/reco:12341234-1234-1234-1234-123412341234'
        )

    def test_property_slug(self):
        self.assertEqual(self.version_1.slug, f'reco_{str(self.version_1.id)[0:6]}')

    def test_property_slug_explicit(self):
        version_2 = WorkerVersion.objects.create(
            id="01234567-89ab-cdef-fedc-ba9876543210",
            worker=self.worker_1,
            revision=self.rev2,
            configuration={"test": "test1"}
        )
        self.assertEqual(version_2.slug, 'reco_012345')

    def test_property_docker_command_empty(self):
        self.assertEqual(self.version_1.docker_command, None)

    def test_property_docker_command_not_empty(self):
        self.version_1.configuration = {'docker': {'command': 'mysupercommand'}}
        self.assertEqual(self.version_1.docker_command, 'mysupercommand')
        self.version_1.configuration = {"test": "test1"}

    def _assert_corpus_worker_version_list(self, response):
        self.assertDictEqual(response.json(), {
            'count': 1,
            'next': None,
            'number': 1,
            'previous': None,
            'results': [{
                'id': str(self.version_1.id),
                'configuration': {'test': 42},
                'revision': {
                    'id': str(self.version_1.revision_id),
                    'hash': '1337',
                    'author': 'Test user',
                    'message': 'My w0rk3r',
                    'created': '2020-02-02T01:23:45.678000Z',
                    'commit_url': 'http://my_repo.fake/workers/worker/commit/1337',
                    'refs': []
                },
                'docker_image': str(self.version_1.docker_image_id),
                'docker_image_iid': None,
                'docker_image_name': self.version_1.docker_image_name,
                'state': 'available',
                'element_count': 9,
                'worker': {
                    'id': str(self.worker_1.id),
                    'name': self.worker_1.name,
                    'type': self.worker_1.type,
                }
            }]
        })

    def test_corpus_worker_version_no_login(self):
        self.corpus.elements.filter(type__slug='word').update(worker_version=self.version_1)

        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self._assert_corpus_worker_version_list(response)

    def test_corpus_worker_version_not_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        self.corpus.elements.filter(type__slug='word').update(worker_version=self.version_1)

        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self._assert_corpus_worker_version_list(response)

    def test_corpus_worker_version_list(self):
        self.client.force_login(self.user)
        self.corpus.elements.filter(type__slug='word').update(worker_version=self.version_1)

        with self.assertNumQueries(11):
            response = self.client.get(reverse('api:corpus-versions', kwargs={'pk': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self._assert_corpus_worker_version_list(response)
