import yaml
from django.test import override_settings
from rest_framework import status
from rest_framework.reverse import reverse

from arkindex.dataimport.models import DataImport, DataImportMode, RepositoryType, WorkerVersion
from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role
from ponos.models import State, Workflow

RECIPE = '''
tasks:
  first:
    image: alpine
'''


class TestWorkflows(FixtureAPITestCase):
    """
    Test workflows from Arkindex existing elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume = Element.objects.get(name='Volume 1')
        cls.pages = Element.objects.get_descending(cls.volume.id).filter(type__slug='page', zone__isnull=False)
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.private_volume = cls.private_corpus.elements.create(
            type=cls.private_corpus.types.create(slug='volume', folder=True),
            name='Hidden',
        )

        # Workers DataImportMode
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(type=RepositoryType.Worker)
        cls.rev_1 = cls.repo.revisions.get()
        cls.rev_2 = cls.repo.revisions.create(
            hash='2',
            message='commit message',
            author='bob',
        )
        cls.version_1 = WorkerVersion.objects.get(worker__slug='reco')
        cls.worker_1 = cls.version_1.worker
        cls.version_2 = WorkerVersion.objects.get(worker__slug='dla')
        cls.worker_2 = cls.version_2.worker

    @override_settings(
        ARKINDEX_TASKS_IMAGE='tasks:latest',
        PONOS_RECIPE={'env': {'SOME_VAR': 'somevalue'}}
    )
    def test_page_workflow(self):
        self.client.force_login(self.user)

        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'type': 'page',
            },
            format='json',
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertIsNone(dataimport.workflow)

        self.assertDictEqual(data, {
            'name': None,
            'id': str(dataimport.id),
            'state': 'unscheduled',
            'corpus': str(self.volume.corpus.id),
            'files': [],
            'mode': 'workers',
            'revision': dataimport.revision,
            'workflow': None,
            'folder_type': None,
            'element_type': 'page',
            'element': None,
            'load_children': False,
            'name_contains': None
        })

        # The process needs to be started to produce a workflow
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        self.assertIsNotNone(dataimport.workflow)

        self.assertEqual(dataimport.workflow.recipes['import'].image, 'tasks:latest')
        self.assertEqual(
            dataimport.workflow.recipes['import'].command,
            'python -m arkindex_tasks.init_elements '
            f'{dataimport.id} '
            '--chunks-number 1'
        )

    def test_volume_workflow(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(self.volume.id),
                'load_children': True
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertIsNone(dataimport.workflow)

        self.assertEqual(dataimport.element, self.volume)
        self.assertEqual(data.pop('element')['id'], str(self.volume.id))
        self.assertDictEqual(data, {
            'name': None,
            'id': str(dataimport.id),
            'state': 'unscheduled',
            'mode': 'workers',
            'corpus': str(self.corpus.id),
            'files': [],
            'revision': dataimport.revision,
            'workflow': None,
            'folder_type': None,
            'element_type': None,
            'load_children': True,
            'name_contains': None
        })

    def test_single_page_workflow(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.element, self.pages.first())

    def test_create_element_workflow_requires_login(self):
        response = self.client.post(reverse('api:corpus-workflow'), {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_element_workflow_required_fields(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus-workflow'), {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'corpus': ['This field is required.']}
        )

    def test_create_element_workflow_requires_corpus_admin(self):
        """
        Only an admin of the target corpus can create a workers process
        """
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Contributor.value)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'name': 'Test',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'corpus': ['You do not have an admin access to this corpus.']})

    def test_create_element_workflow_non_readable_corpus(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.private_corpus.id),
                'name': 'Test',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'corpus': ['Corpus with this ID does not exist.']})

    def test_create_element_workflow_name_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'name': self.pages.first().name[2:5],
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.name_contains, self.pages.first().name[2:5])

    def test_create_element_workflow_name_filter_max_length(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'name': 'A' * 500,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'name': ['Ensure this field has no more than 250 characters.']})

    def test_create_element_workflow_type_filter(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'type': 'page',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.element_type.slug, 'page')

    def test_create_element_workflow_empty_queryset(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'type': 'volume',
                'name': 'ahahah',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['No element match filtering']
        )

    def test_create_element_workflow_empty_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'selection': True,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['No element match filtering']
        )

    def test_create_element_workflow_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'type': 'arachnid',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('type', data)
        self.assertListEqual(
            data['type'],
            ['Unrecognized element type filter']
        )

    def test_create_element_workflow_mixed_element_filters(self):
        """
        Element parameter cannot be set with filtering parameters
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
                'type': 'page',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Filtering parameters (name, type, best_class) cannot be used with element parameter']}
        )

    def test_create_element_workflow_non_folder_no_zone(self):
        self.client.force_login(self.user)
        element = self.corpus.elements.create(type=self.pages.first().type, name='Kill me please')
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(element.id),
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element': ['Element has to be a folder or an element with a zone.']}
        )

    def test_create_element_workflow_mixed_selection_filters(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'element': str(self.volume.id),
                'selection': True,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Filtering parameters (element, name, type, best_class) cannot be used with selection parameter']}
        )

    def test_create_element_workflow_selection_filter(self):
        self.client.force_login(self.user)
        page = self.pages.first()
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [page.id]},
            format='json'
        )
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'selection': True,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertIsNone(dataimport.workflow)
        self.assertEqual(dataimport.elements.get(), page)

        # The process needs to be started to produce a workflow
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        self.assertIsNotNone(dataimport.workflow)

        self.assertEqual(
            dataimport.workflow.recipes['import'].command,
            'python -m arkindex_tasks.init_elements '
            f'{dataimport.id} '
            '--chunks-number 1'
        )

        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(page.id))

    @override_settings(ARKINDEX_FEATURES={'selection': False})
    def test_create_element_workflow_selection_no_selection(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'selection': True,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'selection': ['Selection is not available on this instance.']})

    def test_create_element_workflow_process_name(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'process_name': 'blah',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.name, 'blah')

    def test_create_element_workflow_process_name_max_length(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'process_name': 'A' * 500,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'process_name': ['Ensure this field has no more than 150 characters.']})

    def test_thumbnails_generation_only(self):
        """
        Generating thumbnails without any worker must generate an import task
        and a thumbnails task, when the process is started
        """
        self.client.force_login(self.user)
        corpus_type = self.corpus.types.first()
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'corpus': str(self.corpus.id),
                'type': corpus_type.slug,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertIsNone(dataimport.workflow)

        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport.id)}),
            {
                "chunks": 3,
                'thumbnails': True,
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        self.assertIsNotNone(dataimport.workflow)

        self.assertEqual(
            dataimport.workflow.recipes['import'].command,
            'python -m arkindex_tasks.init_elements '
            f'{dataimport.id} '
            '--chunks-number 3'
        )
        for i in range(1, 4):
            self.assertEqual(
                dataimport.workflow.recipes[f'thumbnails_{i}'].command,
                f'python3 -m arkindex_tasks.generate_thumbnails /data/import/elements_chunk_{i}.json'
            )

    def test_workers_no_worker_runs(self):
        dataimport_2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)

        self.assertIsNone(dataimport_2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport_2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(dataimport_2.id))
        dataimport_2.refresh_from_db()
        self.assertEqual(dataimport_2.state, State.Unscheduled)
        self.assertIsNotNone(dataimport_2.workflow)
        self.maxDiff = None
        self.assertDictEqual(yaml.safe_load(dataimport_2.workflow.recipe)['tasks'], {
            'import':
                {
                    'command': 'python -m arkindex_tasks.init_elements '
                               f'{dataimport_2.id} --chunks-number 1',
                    'image': 'registry.gitlab.com/arkindex/tasks'
                }
        })

    def test_workers_multiple_worker_runs(self):
        dataimport_2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        run_1 = dataimport_2.worker_runs.create(
            version=self.version_1,
            parents=[],
        )
        dataimport_2.worker_runs.create(
            version=self.version_2,
            parents=[run_1.id],
        )

        workflow_tmp = Workflow.objects.create(recipe=RECIPE)
        workflow_tmp.start()

        self.assertIsNone(dataimport_2.workflow)

        self.client.force_login(self.user)
        with self.assertNumQueries(39):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(dataimport_2.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(dataimport_2.id))
        dataimport_2.refresh_from_db()
        self.assertEqual(dataimport_2.state, State.Unscheduled)
        self.assertIsNotNone(dataimport_2.workflow)
        self.maxDiff = None
        self.assertDictEqual(yaml.safe_load(dataimport_2.workflow.recipe)['tasks'], {
            'import':
                {
                    'command': 'python -m arkindex_tasks.init_elements '
                               f'{dataimport_2.id} --chunks-number 1',
                    'image': 'registry.gitlab.com/arkindex/tasks'
                },
            f'reco_{str(self.version_1.id)[0:6]}':
                {
                    'command': None,
                    'image': f'my_repo.fake/workers/worker/reco:{self.version_1.id}',
                    'artifact': str(self.version_1.docker_image.id),
                    'parents': ['import'],
                    'env': {'TASK_ELEMENTS': '/data/import/elements.json', 'WORKER_VERSION_ID': str(self.version_1.id)}
                },
            f'dla_{str(self.version_2.id)[0:6]}':
                {
                    'command': None,
                    'image': f'my_repo.fake/workers/worker/dla:{self.version_2.id}',
                    'artifact': str(self.version_1.docker_image.id),
                    'parents': [f'reco_{str(self.version_1.id)[0:6]}'],
                    'env': {'TASK_ELEMENTS': '/data/import/elements.json', 'WORKER_VERSION_ID': str(self.version_2.id)}
                }
        })
