from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework.exceptions import PermissionDenied
from rest_framework.generics import RetrieveAPIView

from arkindex.documents.models import Corpus, Element
from arkindex.documents.search import search_transcriptions_filter_post
from arkindex.documents.serializers.iiif import (
    ElementAnnotationListSerializer,
    FolderManifestSerializer,
    TranscriptionSearchAnnotationListSerializer,
)
from arkindex.documents.serializers.search import IIIFSearchQuerySerializer
from arkindex.project.elastic import ESTranscription
from arkindex.project.mixins import SearchAPIMixin


class FolderManifest(RetrieveAPIView):
    """
    Retrieve an IIIF manifest for a folder element.
    """
    serializer_class = FolderManifestSerializer

    def get_queryset(self):
        return Element.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
            type__folder=True,
        )

    @method_decorator(cache_page(3600))
    @extend_schema(
        operation_id='RetrieveFolderManifest',
        responses={200: {'type': 'object'}},
        tags=['iiif'],
    )
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class ElementAnnotationList(RetrieveAPIView):
    """
    Retrieve an IIIF annotation list for transcriptions on elements with zones
    """

    serializer_class = ElementAnnotationListSerializer

    def get_queryset(self):
        return Element.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
            zone__isnull=False,
        )

    @method_decorator(cache_page(3600))
    @extend_schema(
        operation_id='RetrieveElementAnnotationList',
        responses={200: {'type': 'object'}},
        tags=['iiif'],
    )
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


@extend_schema_view(
    get=extend_schema(
        operation_id='SearchTranscriptionsAnnotationList',
        responses={200: {'type': 'object'}},
        tags=['iiif'],
    )
)
class TranscriptionSearchAnnotationList(SearchAPIMixin, RetrieveAPIView):
    """
    Retrieve an IIIF Search API annotation list for transcriptions on a folder element
    """

    serializer_class = TranscriptionSearchAnnotationListSerializer
    query_serializer_class = IIIFSearchQuerySerializer
    # For OpenAPI type discovery: an element's ID is in the path
    queryset = Element.objects.none()
    elt = None

    def get_element(self):
        if not self.elt:
            self.elt = Element.objects.get(id=self.kwargs['pk'])
            if self.elt.corpus not in Corpus.objects.readable(self.request.user):
                raise PermissionDenied
        return self.elt

    def get_object(self):
        return self.get_queryset()

    def get_search(self, query=None, min_confidence=0.0, **kwargs):
        return ESTranscription.search() \
            .filter('match', corpus=str(self.get_element().corpus_id)) \
            .filter(
                'terms',
                element=list(
                    Element.objects
                           .filter(
                               paths__path__last=self.get_element().id,
                               type__folder=False,
                               zone__isnull=False,
                           ).values_list('id', flat=True)
                )) \
            .filter('range', score={'gte': min_confidence}) \
            .query('match', text=query)

    def post_process(self, *args):
        return search_transcriptions_filter_post(*args, element_id=self.get_element().id)
