from django.conf import settings
from django.db.models import CharField, F, OuterRef, Prefetch, Subquery, Value
from django.db.models.functions import Ceil, Concat
from rest_framework import serializers

from arkindex.documents.models import Classification, Element, MetaData, MetaType
from arkindex.images.models import Image, Zone
from arkindex.project.tools import bounding_box, build_absolute_url


class ImageResourceManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an image into a IIIF resource.
    """

    def to_representation(self, image):
        assert isinstance(image, Image)
        return {
            "@id": image.get_thumbnail_url(max_width=None, max_height=None),
            "@type": "dctypes:Image",
            "height": image.height,
            "width": image.width,
            "service": {
                "@context": settings.IIIF_IMAGE_CONTEXT,
                "@id": image.url,
                "profile": settings.IIIF_IMAGE_SERVICE_PROFILE
            }
        }


class ImageThumbnailManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an image into a IIIF manifest thumbnail
    """

    def to_representation(self, image):
        assert isinstance(image, Image)
        return {
            "@id": image.get_thumbnail_url(),
            "service": {
                "@context": settings.IIIF_IMAGE_CONTEXT,
                "@id": image.url,
                "profile": settings.IIIF_IMAGE_SERVICE_PROFILE
            }
        }


class ManifestMetadataSerializer(serializers.ModelSerializer):

    label = serializers.CharField(source='name')

    class Meta:
        model = MetaData
        fields = ('label', 'value')


class ElementCanvasManifestSerializer(serializers.BaseSerializer):
    """
    Serialize an element's zone into a IIIF canvas
    """

    def get_other_content(self, element):
        return []

    def to_representation(self, element):
        assert isinstance(element, Element)
        zone = element.zone
        assert isinstance(zone, Zone)
        assert 'request' in self.context, "A request is required to generate absolute URLs"

        if element.folio_metadatas:
            label = element.folio_metadatas[0].value
        else:
            label = element.name

        if element.best_classification:
            label = '{} ; {}'.format(label, element.best_classification)

        _, _, width, height = bounding_box(zone.polygon)

        return {
            "@id": build_absolute_url(element, self.context['request'], 'api:iiif-canvas'),
            "@type": "sc:Canvas",
            "label": label,
            "height": int(height),
            "width": int(width),
            "images": [
                {
                    "@type": "oa:Annotation",
                    "resource": ImageResourceManifestSerializer(zone.image, context=self.context).data,
                    "on": build_absolute_url(element, self.context['request'], 'api:iiif-canvas'),
                    "motivation": "sc:painting"
                }
            ],
            "otherContent": self.get_other_content(element),
        }


class PageCanvasManifestSerializer(ElementCanvasManifestSerializer):
    """
    Serialize a page's zone into a IIIF canvas
    """

    def get_other_content(self, page):
        return [
            {
                "@id": build_absolute_url(page, self.context['request'], 'api:element-annotation-list'),
                "@type": "sc:AnnotationList",
                "label": "Transcriptions",
            }
        ]


class FolderManifestSerializer(serializers.Serializer):
    """
    Serialize a folder into a IIIF manifest
    """

    canvas_serializer = PageCanvasManifestSerializer
    id_url_name = 'api:folder-manifest'

    def to_representation(self, element):
        assert isinstance(element, Element)
        assert 'request' in self.context, "A request is required to generate absolute URLs"

        canvas_elements = self.get_canvases(element)
        canvases = self.canvas_serializer(
            canvas_elements,
            context=self.context,
            many=True
        ).data
        if len(canvas_elements) > 0:
            thumbnail = ImageThumbnailManifestSerializer(canvas_elements[0].zone.image).data
        else:
            thumbnail = None

        services = []
        if settings.ARKINDEX_FEATURES['search']:
            services.append({
                "@context": settings.IIIF_SEARCH_CONTEXT,
                "@id": build_absolute_url(element, self.context['request'], 'api:iiif-search'),
                "profile": settings.IIIF_SEARCH_SERVICE_PROFILE,
                "label": "Search transcriptions",
            })

        return {
            "@context": settings.IIIF_PRESENTATION_CONTEXT,
            "@id": build_absolute_url(element, self.context['request'], self.id_url_name),
            "@type": "sc:Manifest",
            "thumbnail": thumbnail,
            "related": [],
            "structures": self.get_structures(element, canvases),
            "description": "",
            "sequences": [
                {
                    "canvases": canvases,
                    "label": "",
                    "@id": build_absolute_url(element, self.context['request'], 'api:iiif-sequence'),
                    "@type": "sc:Sequence"
                }
            ],
            "viewingHint": "individuals",
            "label": element.name,
            "viewingDirection": "left-to-right",
            "service": services,
            "metadata": ManifestMetadataSerializer(
                element.metadatas.exclude(type=MetaType.HTML),
                context=self.context,
                many=True,
            ).data,
        }

    def get_canvases(self, element):
        assert isinstance(element, Element) and element.type.folder

        # To get a single classification for each page in a element, we force Django to use a subquery;
        # however this only allows fetching a single column. We circumvent this limit using some
        # database functions to get both the MLClass name and the confidence at once.
        # See https://docs.djangoproject.com/en/2.2/ref/models/database-functions/
        # and https://docs.djangoproject.com/en/2.2/ref/models/expressions/#subquery-expressions
        best_classification_subquery = Classification \
            .objects \
            .filter(element=OuterRef('pk')) \
            .order_by('-confidence') \
            .values(label=Concat(
                'ml_class__name',
                Value(' ('),
                Ceil(F('confidence') * 100),
                Value('%)'),
                output_field=CharField(),
            ))[:1]

        folio_prefetch = Prefetch(
            'metadatas',
            MetaData.objects.filter(name='folio'),
            to_attr='folio_metadatas',
        )

        return Element.objects \
            .filter(
                paths__path__last=element.id,  # Direct children
                type__folder=False,
                zone__isnull=False,
            ).annotate(best_classification=Subquery(best_classification_subquery)) \
            .order_by('paths__ordering') \
            .prefetch_related('zone__image__server', folio_prefetch)

    def get_structures(self, element, canvases):
        return [{
            "viewingHint": "top",
            "label": element.name,
            "@id": build_absolute_url(element, self.context['request'], self.id_url_name),
            "ranges": [c['@id'] for c in canvases],
            "@type": "sc:Range"
        }] + [
            {
                "canvases": [c['@id']],
                "label": c['label'],
                "@id": c['@id'],
                "within": c['@id'],
                "@type": "sc:Range"
            }
            for c in canvases]
