from pathlib import Path

import yaml
from drf_spectacular.drainage import error, warn
from drf_spectacular.settings import spectacular_settings

PATCH_FILE = Path(__file__).absolute().parent / 'patch.yml'


def apply_patch_yaml(result, **kwargs):
    """
    drf-spectacular post-processing hook that applies the custom patches from patch.yml.
    """
    with PATCH_FILE.open() as f:
        patch = yaml.safe_load(f)

    for path_name, methods in patch.items():
        if path_name not in result['paths']:
            error(f'OpenAPI path {path_name} does not exist')
            continue

        path_object = result['paths'][path_name]

        for method, operation_patch in methods.items():
            if method not in path_object:
                error(f'Method {method} on OpenAPI path {path_name} does not exist')
                continue

            operation_object = path_object[method]

            # Update responses separately to allow adding status codes without replacing all of them
            if 'responses' in operation_patch and 'responses' in operation_object:
                operation_object['responses'].update(operation_patch.pop('responses'))

            operation_object.update(operation_patch)

    return result


def operation_id_pascal_case(result, **kwargs):
    """
    Transform camelCased operation IDs into PascalCase
    """
    for methods in result['paths'].values():
        for method in methods.values():
            # Ensure there are no 'weird' operation IDs; for example, Spectacular can default to
            # snake_cased operation IDs, or cause duplicated operation IDs to have number suffixes
            if not method['operationId'].isalpha():
                warn(f'Non-alphabetic characters found in operation ID {method["operationId"]}')

            method['operationId'] = method['operationId'][0].upper() + method['operationId'][1:]
    return result


def check_tags(result, **kwargs):
    """
    Check endpoint tags to require them to be defined in settings,
    and require all endpoints to have a tag
    """
    known_tags = set(tag['name'] for tag in spectacular_settings.TAGS)
    for methods in result['paths'].values():
        for method in methods.values():
            tags = method.get('tags')

            if not tags:
                warn(f'Endpoint {method["operationId"]} does not have any assigned tags')
                continue

            unknown_tags = set(tags) - known_tags
            if unknown_tags:
                warn(f'Endpoint {method["operationId"]} uses tags not defined in the project settings: {unknown_tags}')

    return result


def request_body_x_name(result, **kwargs):
    """
    Adds a `x-name` on request bodies to allow them to be used in APIStar
    with the `body={}` keyword argument
    """
    for methods in result['paths'].values():
        for method in methods.values():
            if method.get('requestBody'):
                method['requestBody']['x-name'] = 'body'

    return result
