from unittest import TestCase
from unittest.mock import call, patch

from arkindex.project.openapi import apply_patch_yaml, check_tags, operation_id_pascal_case, request_body_x_name


@patch('arkindex.project.openapi.hooks.warn')
@patch('arkindex.project.openapi.hooks.error')
class TestSpectacularHooks(TestCase):

    @patch('arkindex.project.openapi.hooks.yaml.safe_load')
    def test_patch_yaml(self, yaml_mock, error_mock, warn_mock):
        """
        Test the patch_yaml post-processing hook that should override existing
        operation objects and cause errors for non-existent operations.
        It should also extend, instead of override, the responses to allow adding
        response codes without removing the existing ones.
        """
        schema = {
            'paths': {
                '/a': {
                    'get': {
                        'operationId': 'RetrieveA',
                        'responses': {
                            '200': {
                                'type': 'object'
                            }
                        }
                    }
                }
            }
        }

        yaml_mock.return_value = {
            '/a': {
                'get': {
                    'operationId': 'ListA',
                    'tags': ['a'],
                    # Should extend, not override responses
                    'responses': {
                        '418': {
                            'type': 'string'
                        }
                    }
                },
                # Non-existent method
                'post': {
                    'operationId': 'CreateA',
                }
            },
            # Non-existent path
            '/b': {
                'get': {
                    'operationId': 'RetrieveB',
                }
            }
        }

        schema = apply_patch_yaml(schema)

        self.assertDictEqual(schema, {
            'paths': {
                '/a': {
                    'get': {
                        'operationId': 'ListA',
                        'tags': ['a'],
                        'responses': {
                            '200': {
                                'type': 'object'
                            },
                            '418': {
                                'type': 'string'
                            }
                        }
                    }
                }
            }
        })

        self.assertListEqual(error_mock.call_args_list, [
            call('Method post on OpenAPI path /a does not exist'),
            call('OpenAPI path /b does not exist'),
        ])
        self.assertFalse(warn_mock.called)

    def test_operation_id_pascal_case(self, error_mock, warn_mock):
        """
        Test the operation_id_pascal_case post-processing hook that should
        turn camelCase operation IDs generated by Spectacular into PascalCase
        and warn about non-alphabetic IDs, as duplicate IDs can be resolved
        by Spectacular with numeric suffixes.
        """
        schema = operation_id_pascal_case({
            'paths': {
                '/a': {
                    'get': {
                        'operationId': 'retrieveA',
                    },
                    'post': {
                        'operationId': 'CreateA',
                    },
                    'put': {
                        'operationId': 'createA2',
                    }
                }
            }
        })
        self.assertDictEqual(schema, {
            'paths': {
                '/a': {
                    'get': {
                        'operationId': 'RetrieveA',
                    },
                    'post': {
                        'operationId': 'CreateA',
                    },
                    'put': {
                        'operationId': 'CreateA2',
                    }
                }
            }
        })
        self.assertFalse(error_mock.called)
        self.assertEqual(warn_mock.call_count, 1)
        self.assertEqual(warn_mock.call_args, call(
            'Non-alphabetic characters found in operation ID createA2'
        ))

    @patch('arkindex.project.openapi.hooks.spectacular_settings.TAGS', [{'name': 'a'}])
    def test_check_tags(self, error_mock, warn_mock):
        """
        Test the check_tags post-processing hook that should ensure each
        endpoint has at least one tag and only uses tags defined in
        arkindex.project.settings.
        """
        check_tags({
            'paths': {
                '/a': {
                    'get': {
                        'operationId': 'RetrieveA',
                    },
                    'post': {
                        'operationId': 'CreateA',
                        'tags': ['a'],
                    },
                    'delete': {
                        'operationId': 'DestroyA',
                        'tags': ['a', 'b'],
                    }
                }
            }
        })
        self.assertFalse(error_mock.called)
        self.assertListEqual(warn_mock.call_args_list, [
            call('Endpoint RetrieveA does not have any assigned tags'),
            call("Endpoint DestroyA uses tags not defined in the project settings: {'b'}")
        ])

    def test_request_body_x_name(self, error_mock, warn_mock):
        """
        Test the request_body_x_name post-processing hook adds x-name: body to all request bodies
        """
        schema = request_body_x_name({
            'paths': {
                '/a': {
                    'post': {
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {'type': 'object'}
                                }
                            }
                        }
                    },
                    'delete': {
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {'type': 'object'}
                                }
                            }
                        }
                    }
                }
            }
        })
        self.assertDictEqual(schema, {
            'paths': {
                '/a': {
                    'post': {
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {'type': 'object'}
                                }
                            },
                            'x-name': 'body'
                        }
                    },
                    'delete': {
                        'requestBody': {
                            'content': {
                                'application/json': {
                                    'schema': {'type': 'object'}
                                }
                            },
                            'x-name': 'body'
                        }
                    }
                }
            }
        })
