import tempfile
from unittest.mock import patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import DataImport, DataImportMode
from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User
from ponos.authentication import AgentUser
from ponos.models import Agent, Artifact, Farm, Secret, State, encrypt


@override_settings(PONOS_PRIVATE_KEY='staging')
class TestPonosView(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.dataimport_corpus = Corpus.objects.create(name='Another public corpus', public=True)
        cls.corpus_admin = User.objects.create(email='corpusadmin@test.me')
        cls.corpus_admin.rights.create(content_object=cls.dataimport_corpus, level=Role.Admin.value)

        cls.creator = User.objects.create(email="creator@user.me")
        cls.artifact = Artifact.objects.get(path='/path/to/docker_build')
        cls.task = cls.artifact.task
        cls.workflow = cls.task.workflow
        cls.task = cls.workflow.tasks.get()
        cls.dataimport = DataImport.objects.create(
            mode=DataImportMode.Repository,
            creator=cls.creator,
            workflow=cls.workflow,
            corpus=cls.dataimport_corpus
        )

    @patch("ponos.models.settings")
    def test_internal_user_retrieve_secret(self, settings_mock):
        """
        Only Arkindex internal users should be able to retrieve a Ponos secret
        """
        _, path = tempfile.mkstemp()
        settings_mock.PONOS_PRIVATE_KEY = path
        with open(path, "wb") as f:
            f.write(b"pikachu")
        account_name = "bank_account/1337/A"
        secret = Secret.objects.create(
            name=account_name,
            nonce=b"1337" * 4,
            content=encrypt(b"1337" * 4, "1337$"),
        )
        internal_user = User.objects.create(email='internal', is_internal=True)
        self.client.force_login(internal_user)
        response = self.client.get(reverse('secret-details', kwargs={'name': account_name}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'id': str(secret.id),
            'name': account_name,
            'content': '1337$'
        })

    def test_non_internal_user_cannot_retrieve_secret(self):
        non_internal_users = User.objects.bulk_create(User(**args) for args in (
            {'email': 'admin', 'is_admin': True},
            {'email': 'active', 'is_active': True},
            {'email': 'unverified', 'verified_email': True},
            {'email': 'anonymous'}
        ))
        for user in non_internal_users:
            self.client.force_login(user)
            response = self.client.get(reverse('secret-details', kwargs={"name": "secret/name"}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_agents(self):
        """
        Only authenticated users should have the ability to list agents
        """
        response = self.client.get(reverse('ponos-agents'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.client.force_login(User.objects.create())
        response = self.client.get(reverse('ponos-agents'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_retrieve_agent(self):
        """
        Only authenticated users should have the ability to retrieve details of an agent
        """
        agent = Agent.objects.create(
            cpu_cores=3,
            cpu_frequency=3e9,
            farm_id=Farm.objects.create().id,
            ram_total=2e9,
            last_ping='1999-09-09'
        )

        response = self.client.get(reverse('ponos-agent-details', kwargs={'pk': str(agent.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.client.force_login(User.objects.create())
        response = self.client.get(reverse('ponos-agent-details', kwargs={'pk': str(agent.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_update_task(self):
        """
        Only users with an admin privilege and creators have the ability to update a dataimport task
        """
        test_cases = (
            (None, status.HTTP_403_FORBIDDEN, 1),
            (self.creator, status.HTTP_200_OK, 9),
            (self.user, status.HTTP_403_FORBIDDEN, 9),
            (self.superuser, status.HTTP_200_OK, 7),
            (self.corpus_admin, status.HTTP_200_OK, 11),
        )
        for user, status_code, requests_count in test_cases:
            if user:
                self.client.force_login(user)
            with self.assertNumQueries(requests_count):
                response = self.client.patch(
                    reverse('ponos-task-update', kwargs={'pk': str(self.task.id)}),
                    json={'state': 'stopping'}
                )
                self.assertEqual(response.status_code, status_code)

    def test_download_artifacts(self):
        """
        Only users with an admin privilege and creators have the ability to
        download an artifact of a dataimport task
        """
        test_cases = (
            (None, status.HTTP_403_FORBIDDEN, 2),
            (self.creator, status.HTTP_302_FOUND, 6),
            (self.user, status.HTTP_403_FORBIDDEN, 9),
            (self.superuser, status.HTTP_302_FOUND, 4),
            (self.corpus_admin, status.HTTP_302_FOUND, 8),
        )
        for user, status_code, requests_count in test_cases:
            if user:
                self.client.force_login(user)
            with self.assertNumQueries(requests_count):
                response = self.client.get(
                    reverse('ponos-artifact-dl', kwargs={'pk': str(self.task.id), 'path': self.artifact.path}),
                    follow=False
                )
                self.assertEqual(response.status_code, status_code)

    def test_download_artifacts_by_agent(self):
        """
        Agents should still be able to download artifacts in order
        to run followup tasks
        """
        agent_user = AgentUser.objects.create(
            cpu_cores=3,
            cpu_frequency=3e9,
            farm_id=Farm.objects.create().id,
            ram_total=2e9,
            last_ping='1999-09-09'
        )

        with self.assertNumQueries(3):
            response = self.client.get(
                reverse('ponos-artifact-dl', kwargs={'pk': str(self.task.id), 'path': self.artifact.path}),
                follow=False,
                HTTP_AUTHORIZATION="Bearer {}".format(agent_user.token.access_token),
            )
            self.assertEqual(response.status_code, status.HTTP_302_FOUND)

    def test_retrieve_workflow(self):
        """
        A user with a guest right on a process cannot retrieve its workflow
        """
        self.client.force_login(self.user)
        self.dataimport_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        with self.assertNumQueries(9):
            response = self.client.get(reverse('ponos-workflow-details', kwargs={'pk': str(self.workflow.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_retry_workflow_requires_admin_access(self):
        """
        A user cannot retry a workflow if they have no admin access to their process
        """
        self.task.state = State.Running
        self.task.save()
        self.client.force_login(self.user)
        self.dataimport_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse('ponos-workflow-details', kwargs={'pk': str(self.workflow.id)}),
                {'status': 'stopped'}
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(response.json(), {'detail': 'You do not have an admin access to this process.'})

    def test_retry_workflow(self):
        self.task.state = State.Running
        self.task.save()
        self.client.force_login(self.user)
        self.dataimport_corpus.memberships.create(user=self.user, level=Role.Admin.value)
        with self.assertNumQueries(14):
            response = self.client.patch(
                reverse('ponos-workflow-details', kwargs={'pk': str(self.workflow.id)}),
                {'status': 'stopped'}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
