"""
Helper methods to trigger tasks in asynchronous workers
"""
from typing import Optional, Union
from uuid import UUID

from django.conf import settings

from arkindex.documents import tasks
from arkindex.documents.managers import ElementQuerySet
from arkindex.documents.models import Corpus, Element, Entity


def reindex_start(*,
                  corpus: Union[Corpus, UUID, str] = None,
                  element: Union[Element, UUID, str] = None,
                  entity: Union[Entity, UUID, str] = None,
                  transcriptions: bool = False,
                  elements: bool = False,
                  entities: bool = False,
                  drop: bool = False,
                  user_id: Optional[int] = None) -> None:
    """
    Reindex elements into ElasticSearch.

    One of `corpus`, `element` or `entity` may be specified.
    Otherwise all elements will be picked.
    If `element` is specified, its children will be included.

    `transcriptions`, `elements` and `entities` booleans represents Elasticsearch corresponding indexes.
    Update will be run on indexes set to True only.

    `drop` will REMOVE all indexes entirely and recreate them before indexing.
    This is only allowed when reindexing all elements.
    """
    if not settings.ARKINDEX_FEATURES['search']:
        return

    element_id = None
    corpus_id = None
    entity_id = None
    if isinstance(element, Element):
        element_id = str(element.id)
    elif element:
        element_id = str(element)

    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
    elif corpus:
        corpus_id = str(corpus)

    if isinstance(entity, Entity):
        entity_id = str(entity.id)
    elif entity:
        entity_id = str(entity)

    if element_id:
        description = f'Indexation of element {element_id}'
    elif entity_id:
        description = f'Indexation of entity {entity_id}'
    elif corpus_id:
        description = f'Indexation of corpus {corpus_id}'
    else:
        description = 'Full reindexation'

    tasks.reindex_start.delay(
        element_id=element_id,
        corpus_id=corpus_id,
        entity_id=entity_id,
        transcriptions=transcriptions,
        elements=elements,
        entities=entities,
        drop=drop,
        description=description,
        user_id=user_id,
    )


def corpus_delete(corpus: Union[Corpus, UUID, str], user_id: Optional[int] = None) -> None:
    """
    Delete a whole corpus without killing a server by removing all related
    models first and lowering the amount of foreign keys Django has to handle.
    """
    if isinstance(corpus, Corpus):
        corpus_id = str(corpus.id)
        description = f'Deletion of corpus {corpus.name}'
    else:
        corpus_id = str(corpus)
        description = f'Deletion of corpus {corpus_id}'

    tasks.corpus_delete.delay(corpus_id=corpus_id, description=description, user_id=user_id)


def element_trash(queryset: ElementQuerySet,
                  delete_children: bool = True,
                  user_id: Optional[int] = None) -> None:
    """
    Run ElementQuerySet.trash to delete a batch of elements.
    """
    assert isinstance(queryset, ElementQuerySet), 'Only Element querysets can be trashed'
    tasks.element_trash.delay(
        queryset=queryset,
        delete_children=delete_children,
        user_id=user_id,
        description='Element deletion',
    )
